# meds/bots/__init__.py
#
#

""" Several type of bots. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from meds.utils.parse import txt_parse
from meds.utils.time import short_date
from meds.dispatcher import Dispatcher
from meds.main import main, fleet, plugins
from meds.defines import BLA, ENDC
from meds.object import Object
from meds.errors import error

from meds.errors import RemoteDisconnect, NotImplemented

import logging
import socket
import time

## BOTS

class Bot(Dispatcher):

    """ Basic Bot. """

    default = ""
    cc = ""

    def __init__(self, *args, **kwargs):
        """ 
            __init__()

            Bot initialisation code.

            :param list args: arguments list
            :param dict kwargs: named arguments dict

            >>> bot = Bot()
            >>> bot.json()
            '{"channels": []}'

        """
        Dispatcher.__init__(self, *args, **kwargs)
        self._start = time.time()
        self.channels = []
        if "channel" in kwargs: self.channels.append(kwargs["channel"])

    ## OUTPUT

    def _raw(self, *args, **kwargs):
        """ 
            _raw(txt)

            Output txt with print()

            :param string txt: text to print.
            :return: None

            >>> b = Bot()
            >>> b._raw("bla")
            bla

        """
        print(args[0])

    def say(self, *args, **kwargs):
        """ 
            say(channel, txt)

            Output text to channel

            :param string channel: channel to output to.
            :param string txt: text to output
            :return: None

            >>> b = Bot()
            >>> b.say("#dunkbots", "yo!")
            yo!
 
        """
        if args: self._raw(args[1])
        
    def announce(self, *args, **kwargs):
        """ 
            announce(txt)

            Announce text to channel.

            :param string txt: text to announce.
            :return: None

            >>> b = Bot()
            >>> b.channels = ["#dunbkots",]
            >>> b.announce("okdan")
            okdan

        """
        self._last = time.time()
        logging.info("# %s/announce %s" % (self.type, " ".join(self.channels)))
        for channel in self.channels: self.say(channel, args[0])

    # CHANNELS

    def join(self, *args, **kwargs):
        """ 
            join(channel)

            Join a channel

            :param string channel: channel to join 
            :return: None

            >>> b = Bot()
            >>> b.join("#dunkbots")

        """
        pass
        
    def joining(self, *args, **kwargs):
        """ 
            joining(*args, **kwargs)

            Callback called when channel is joined

            :param list args: arguments list
            :param dict kwargs: named arguments dict
            :return: None

            >>> b = Bot()
            >>> b.join = lambda x: print(x)
            >>> b.single("JOIN #dunkbots :mekker")
            []

        """
        for channel in self.channels: self.join(channel)

    ## LOOP

    def start(self, *args, **kwargs):
        """ 
            start()

            Starts the bot's mainloop

            :return: None

            >>> b = Bot()
            >>> b.start(single=True)

        """
        logging.warn("# %s/start" % self.type)
        self._status = "ok"
        if "collect" in kwargs: logging.warn("# %s/enable collect" % self.type)
        while self._status:
            try:
                event = self.event()
                if not self._status: break
                if not event: continue
                if "txt" in event: event.parsed = txt_parse(self, event.txt)
                else: event.parsed = txt_parse(self, "")
                thrs = self.dispatch(event)
                if thrs and kwargs.get("collect", False):
                    result = self.collect(thrs)
                    logging.info("%s/collect %s" % (self.type, ", ".join([str(x) for x in result])))
                if "single" in kwargs: break
            except RemoteDisconnect: time.sleep(5.0) ; continue
            except (EOFError, KeyboardInterrupt): raise
            except: error()
        self.ready()
        logging.warn("# stop %s" % self.type)

    def event(self, *args, **kwargs):
        """ 
            event(txt)

            Create an event based on given text.

            :param string txt: optional text to create event from.
            :return: Object

            >>> b = Bot()
            >>> e = b.event("ls")
            >>> e.json(sort_keys=True)
            '{"origin": "root@self", "txt": "ls"}'

        """
        event = Object()
        event._target = self
        if args: event.txt = args[0]
        event.origin = "root@self"
        return event

    ## DISPATCH

    def single(self, *args, **kwargs):
        """ 
            single(txt)

            Run a single string on the dispatcher

            :param string txt: text to dispatch
            :return: list of threads started

            >>> b = Bot()
            >>> b.single("PING :yo!")
            []
 
        """
        if not args[0]: return
        event = Object()
        event._target = self
        event.origin = "meds@shell"
        event.txt = args[0]
        event.parsed = txt_parse(self, event.txt)
        return self.dispatch(event)

    def dispatch(self, *args, **kwargs):
        """ 
            dispatch(event)

            Dispatch an event registered on the bot 

            :param Object event: event to dispatch
            :return: list of threads started

            >>> b = Bot()
            >>> ignore = plugins.load("meds.plugins.desk")
            >>> e = b.event("ping")
            >>> b.dispatch(e)
            
        """
        event = args[0]
        if "txt" not in event: return
        if "parsed" not in event: event.parsed = txt_parse(self, event.txt)
        cmnds = self.find(event.parsed.cmnd)
        thrs = []
        for cmnd in cmnds:
            try: o = self.put(cmnd.func, event) ; thrs.append(o)
            except AttributeError: continue
            except: error()
        return thrs

    def find(self, *args, **kwargs):
        """ 
            find(txt)

            Search the bot for a registered command matching the given text

            :param string txt: text to search for
            :return: list of commands found

            >>> b = Bot()
            >>> a = plugins.load("meds.plugins.desk")
            >>> b.find("ls") # doctest: +ELLIPSIS
            [{'func': <function desk_ls at ... >, 'plugname': 'meds.plugins.desk', 'cmnd': 'desk.ls'}]

        """
        from meds.main import main
        cmnd = args[0]
        result = []
        cmnds = main.match(cmnd)
        if not cmnds:
            try: cmnds = main[cmnd]
            except KeyError: cmnds = []
        return cmnds

            
## TESTBOT

class TestBot(Bot):

    """ USE WITH Bot.put() and Bot.read(). """

    testing = True
    cc = "!"
    default = ""
    channels = ["#test", ]

    def event(self, *args, **kwargs):
        """ 
            event(txt)

            Create an event as if originating from this bot

            :param string txt: text to create event from
            :return: Object

            >>> b = TestBot()
            >>> e = b.event("PING :yo!")
            >>> e.json(sort_keys=True)
            '{"loglevel": "warning", "origin": "test@meds", "parsed": {"args": [":yo!"], "cc": "!", "cmnd": "", "default": false, "notwant": {}, "rest": ":yo!", "switch": {}, "txt": "PING :yo!", "want": {}}, "txt": "PING :yo!"}'

        """
        o = Object()
        o.origin = "test@meds"
        o._target = self
        o.loglevel = "warning"
        o.txt = args[0]
        o.parsed = txt_parse(self, o.txt)
        return o

## CMNDS

def join(event):
    """ 
        join(event)

        Callback for the join command

        :param Object event: event passed to command
        :return: None

        >>> b = Bot()
        >>> b.single("!join #dunkbots")
        []

    """
    if not event.parsed.args: return
    channel = event.parsed.args[0]
    for bot in fleet:
        if "join" not in bot: continue
        bot.join(channel)
        event.reply("join %s (%s)" % (channel, bot.server))

main.register("bot.join", join)

def part(event):
    """ 
        part(event)

        Callback for the part command

        :param Object event: event passed to command
        :return: None

        >>> from meds.bots import Bot
        >>> b = Bot()
        >>> b.single("!part #dunkbots")
        []

    """
    if not event.parsed.args: return
    for bot in fleet:
        try: bot.part(event.parsed.args[0])
        except: continue
        event.reply("part %s (%s)" % (channel, bot.server))

main.register("bot.part", part)

def nick(event):
    """ 
        nick(event)

        Callback of the nick command

        :param Object event: event passed to command
        :return: None

        >>> from meds.bots import Bot
        >>> b = Bot()
        >>> b.single("!nick botfather")
        []

    """
    if not event.parsed.args: return
    for bot in fleet:
        try: bot.nick(event.parsed.args[0])
        except: continue
        event.reply("nick %s" % bot.server)

main.register("bot.nick", nick)

def quit(event):
    """ 
        quit(event)

        Callback for the quit command

        :param Object event: event passed to command
        :return: None

        >>> from meds.bots import Bot
        >>> b = Bot()
        >>> b.single("!quit")
        []

    """
    if not event.parsed.args: return
    for bot in fleet:
        try: bot.quit(event.parsed.rest)
        except: continue
        event.reply("quit %s" % bot.server)
    
main.register("bot.quit", quit)
