# meds/bots/irc.py
#
#

""" Internet Relay Chat Bot. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from meds.main import main, fleet, plugins
from meds.dispatcher import Dispatcher
from meds.bots import Bot
from meds.object import Object

from meds.utils.other import split_txt
from meds.utils.parse import txt_parse
from meds.utils.trace import get_location
from meds.errors import RemoteDisconnect, NotConnected, error
from meds.cfg import cfg
from meds import __version__

import logging
import _thread
import socket
import queue
import time
import sys
import re

## KLASSEN

class IRC(Bot):

    """ Internet Relay Chat Bot. """

    default = ""
    marker = "\r\n"
    cc = "!"

    def __init__(self, *args, **kwargs):
        Bot.__init__(self, *args, **kwargs)
        self._lock = _thread.allocate_lock()
        self._outqueue = queue.Queue()
        self._sock = None
        self._buffer = []
        self._lastline = ""
        self._error = ""
        self._status = "started"
        self._handlers = Dispatcher()
        self.add("004", self.connected)
        self.add("ERROR", self.errored)
        self.add("366", self.h366)
        self.add("433", self.h433)
        self.add("451", self.logon)
        self.add("513", self.h513)
        self.add("PING", self.pinged)
        self.add("PONG", self.ponged)
        self.add("QUIT", self.quited)
        self.add("INVITE", self.invited)
        self.add("PRIVMSG", self.privmsged)
        self.add("NOTICE", self.noticed)
        self.add("JOIN", self.joined)
        self.channels = []
        self.encoding = "utf-8"
        if "realname" not in self: self.realname = "meds"
        if "username" not in self: self.username = "meds"
        if "server" not in self: self.server = "localhost"
        if "port" not in self: self.port = 6667
        if "nick" not in self: self.nick = "meds"
        if "channel" in self and self.channel not in self.channels: self.channels.append(self.channel)

    def add(self, *args, **kwargs): self._handlers.register(*args, **kwargs)

    ## HANDLERS

    def joined(self, *args, **kwargs):
        """ 
            joined(event)

            Called when a user joins a channel.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("JOIN #dunkbots :dunker")
            >>> bot.joined(e)

        """
        event = args[0]
        if event.channel not in self.channels: self.channels.append(event.channel)

    def errored(self, *args, **kwargs):
        """ 
            errored(event)

            Called when a user joins a channel.

            :param Object event: event passed onto the callback


            >>> bot = IRC()
            >>> e = bot.event("error occured")
            >>> bot.errored(e)

        """
        event = args[0]
        logging.error("%s/error %s" % (self.type, event.txt))
        self._status = "error"

    def noticed(self, *args, **kwargs):
        """ 
            noticed(event)

            Virtual callback called when a user joins a channel. Needs to be overloaded.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("NOTICE #dunkbots :dunker")
            >>> bot.noticed(e)

        """
        pass

    def pinged(self, *args, **kwargs):
        """ 
            pinged(event)

            Ping callback.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("PING :dunker")
            >>> bot.pinged(e)
            pong

        """
        event = args[0]
        self.pongcheck = True
        self.pong(event.txt)

    def ponged(self, *args, **kwargs):
        """ 
            ponged(event)

            Virtual pong callback. Needs to be overloaded

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("PONG :dunker")
            >>> bot.ponged(e)

        """
        pass 

    def h433(self, *args, **kwargs):
        """ 
            h433(event)

            Handler for 433 messages.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("433 dunker mekker :bla")
            >>> bot.h433(e)

        """
        event = args[0]
        self.donick(event.arguments[1] + "_")

    def quited(self, *args, **kwargs):
        """ 
            quited(event)

            Handler for 433 messages.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("QUIT :dunker")
            >>> bot.quited(e)

        """
        event = args[0]
        if "Ping timeout" in event.txt and event.nick == self.nick: self.connect()

    def h366(self, *args, **kwargs):
        """ 
            h366(event)

            Handler for 366 messages. Needs to be overloaded.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("366 :dunker")
            >>> bot.h366(e)

        """
        pass 
      
    def invited(self, *args, **kwargs):
        """ 
            invited(event)

            Called be bot is invited, does a autojoin.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("INVITE :dunker")
            >>> bot.invited(e)

        """
        event = args[0]
        self.join(event.channel)

    def privmsged(self, *args, **kwargs):
        """ 
            privmsged(event)

            Called be bot is privmsged, dispatches potential command to the bot.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("PRIVMSG meds :!version")
            >>> bot.privmsged(e)

        """
        event = args[0]
        if event.txt.startswith("\001DCC"): self.dccconnect(event) ; return
        elif event.txt.startswith("\001VERSION"): self.ctcpreply(event.nick, "VERSION MEDS #%s - http://pypi.python.org/bart/meds" % __version__) ; return
        Bot.dispatch(self, event)

    def h513(self, *args, **kwargs):
        """ 
            h513(event)

            513 callback, echo's a PONG.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("513 bla mekker miep klaas meds okdan :yo!")
            >>> bot.h513(e)

        """
        event = args[0]
        self.raw("PONG %s" % event.arguments[6])

    def ctcped(self, event):
        """ 
            ctcped(event)

            Callback called when CTCP is received. Virtual method.

            :param Object event: event passed onto the callback

            >>> bot = IRC()
            >>> e = bot.event("CTCP meds :version")
            >>> bot.ctcped(e)

        """
        pass

    def connected(self, *args, **kwargs):
        """ 
            connected()

            Callback called when the bot is connected to the  server.  If self.servermodes is set, it is send to the server.


            >>> bot = IRC()
            >>> bot.connected()
            
        """

        if "servermodes" in self: self.raw("MODE %s %s" % (self.nick, self.servermodes))
        logging.warn("# %s/connected with %s" % (self.type, self.server))
        self.joining()

    ## OUTPUT

    def raw(self, *args, **kwargs):
        """ 
            raw(txt)

            Send a raw string to the server.

            :param str txt: raw text
            :param boolean kwargs["nolog"]: Switch to disable logging
            :raises meds.errors.NotConnected: Exception raised when bot is not connected
 
            >>> bot = IRC()
            >>> bot.raw("PRIVMSG dunker :yo!")

        """
        txt = args[0]
        if not txt.endswith(self.marker): txt += self.marker
        txt = txt[:512]
        txt = bytes(txt, "utf-8")
        if "nolog" not in kwargs: logging.info("> %s/write %s" % (self.type, txt))
        try:
            if 'ssl' in self and self.ssl: self._sock.write(txt)
            else: self._sock.send(txt)
        except BrokenPipeError: self.connect()

    ## CONNECT

    def connect(self):
        """ 
            connect()

            Connect the bot to the server set in self.server, use self.port to connect.


            >>> bot = IRC()
            >>> bot.connect()
            True

        """
        self.stopped = False
        logging.warn("# %s/connecting to %s" % (self.type, self.server))
        if "ipv6" in self: self._oldsock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        else: self._oldsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server = self.bind()
        self._oldsock.settimeout(60)
        self._oldsock.connect((self.server, int(str(self.port or 6667))))
        self.blocking = True
        self._oldsock.setblocking(self.blocking)
        self._oldsock.settimeout(700.0)
        self.fsock = self._oldsock.makefile("r")
        if 'ssl' in self and self['ssl']: self._sock = socket.ssl(self._oldsock)
        else: self._sock = self._oldsock
        return True

    ## DCC

    def dcc(self, event, s):
        """ 
            dcc(event, s)

            Callback called when user is connecting with DCC CHAT

            :param Object event: event passed to callback
            :param socket s: socket used by user to connect to DCC CHAT

            >>> bot = IRC()
            >>> e = bot.event("DCC CHAT meds", nick="dunker")
            >>> s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            >>> bot.dcc(e, s)

        """
        s.send(bytes('Welcome to MEDS ' + event.nick + " !!\n", self.encoding))
        main.put(self.dccloop, event, s)

    def dccloop(self, event, s):
        """ 
            dccloop(event, s)

            :param Object event: event passed to callback
            :param socket s: socket used by user to connect to DCC CHAT

            Threadloop handling DCC CHAT connection.

            >>> from meds.bots.irc import IRC
            >>> bot = IRC()
            >>> e = bot.event("DCC CHAT meds")
            >>> s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            >>> bot.dccloop(e, s)

        """
        sockfile = s.makefile('rw')
        s.setblocking(True)
        while 1:
            try:
                res = sockfile.readline()
                if not res: break
                res = res.rstrip()
                logging.info("< %s/DCC %s" % (self.type, event.origin))
                o = Object()
                o._target = self
                o.txt = res
                o.outer = sockfile
                o.origin = event.origin
                o.parsed = txt_parse(self, res, "none")
                Bot.dispatch(self, o)
            except socket.timeout: time.sleep(0.01)
            except socket.error as ex:
                if ex.errno in [socket.EAGAIN, ]: continue
                else: raise
            except Exception as ex: error()
        sockfile.close()

    def dccconnect(self, event):
        """ 
            dccconnect(event)

            :param Object event: event passed to callback

            Callback to connect back to the user on a DCC CHAT request.

            # todo - print received event so we can create it in doctest to use

            >>> bot = IRC()
            >>> e = bot.event("BLA meds: mekker")
            >>> bot.dccconnect(e)

        """
        try:
            addr = event.parsed.args[2] ; port = event.parsed.args[3][:-1]
            port = int(port)
            if re.search(':', addr): s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            else: s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((addr, port))
        except Exception as ex: error() ; return
        self.dcc(event, s)

    def start(self, *args, **kwargs):
        """ 
            start()

            Start the IRC Bot

            :param boolean single: When set to true, do only one loop.

            >>> bot = IRC()
            >>> bot.start(single=True)

        """
        self._status = "connect"
        self.connecting()
        Bot.start(self, *args, **kwargs)

    def parse(self, *args, **kwargs):
        """ 
            parse(txt)

            Parse a string into an IRC event.

            :param str txt: text to parse.
            :return: Object

            >>> from meds.bots.irc import IRC
            >>> bot = IRC()
            >>> bot.parse("PRIVMSG meds :!version)
           
        """
        rawstr = str(args[0])
        obj = Object()
        obj._target = self
        obj.servermsg = False
        splitted = re.split('\s+', rawstr)
        if not rawstr[0] == ':': obj.servermsg = True
        obj.pre = splitted[0]
        if obj.servermsg: obj.type = obj.pre
        else: obj.type = splitted[1]
        try:
            nickuser = obj.pre.split('!')
            obj.origin = nickuser[1]
            obj.nick = nickuser[0][1:]
        except IndexError: obj.origin = obj.pre ; obj.servermsg = True
        if obj.type in pfc:
            obj.arguments = splitted[2:pfc[obj.type]+2]
            txtsplit = re.split('\s+', rawstr, pfc[obj.type]+2)
            obj.txt = txtsplit[-1]
        else: obj.arguments = splitted[2:]
        if obj.arguments: obj.target = obj.arguments[0]
        obj.postfix = ' '.join(obj.arguments)
        if not "txt" in obj: obj.txt = rawstr.rsplit(":")[-1]
        if obj.txt.startswith(":"): obj.txt = obj.txt[1:]
        if not "channel" in obj:
            for c in obj.arguments + [obj.txt, ]:
                if c.startswith("#"): obj.channel = c
        if obj.servermsg:
            obj.origin = obj.origin[1:-1]
            obj.channel = obj.origin
        if not "origin" in obj: obj.origin = obj.channel
        return obj

    def event(self, *args, **kwargs):
        """ 
            event(txt="")

            Create an IRC event from a string. 

            :param str txt: optional text to parse.
            :return: Object

            >>> bot = IRC()
            >>> bot.parse("PRIVMSG meds :!version")

        """
        if args: line = args[0]
        else:
            if not self._buffer: self.some()
            line = self._buffer.pop(0)
        return self.parse(line.rstrip())

    def dispatch(self, *args, **kwargs):
        """ 
            dispatch(event)

            Dispatch an IRC event on the IRC bot.

            :param Object event: event to dispatch

            >>> bot = IRC()
            >>> e = bot.event("PRIVMSG #dunkbots :!version")
            >>> bot.dispatch(e)

        """
        event = args[0]
        if event.type not in self._handlers: return
        for handler in self._handlers[event.type]: self._handlers.put(handler.func, *args, **kwargs)

   ## INPUT

    def some(self, *args, **kwargs):
        """ 
            some()

            Read some bytes from the server socket.


            >>> bot = IRC()
            >>> bot.some()

        """
        if "ssl" in self and self.ssl: inbytes = self._sock.read()
        else: inbytes = self._sock.recv(512)
        txt = str(inbytes, self.encoding)
        if txt == "": raise RemoteDisconnect()
        self._lastline += txt
        splitted = self._lastline.split(self.marker)
        for s in splitted[:-1]:
            self._buffer.append(s)
            if not s.startswith("PING"): logging.warn("< %s/read %s" % (self.type, s.strip()))
        self._lastline = splitted[-1]

    ## OUTPUT

    def announce(self, *args, **kwargs):
        """ 
            announce(txt)

            Announce txt on all channels this bot is connected to.

            :param str txt: text to announce.

            >>> bot = IRC()
            >>> bot.announce("yo!")

        """
        for channel in self.channels: self._outqueue.put((channel, args[0]))

    def push(self, *args, **kwargs):
        """ 
            push()

            Start the push thread to output queued text to the server.

            >>> bot = IRC()
            >>> bot.push(("#dunkbots", "yo!"))

        """
        kw = Object(**kwargs)
        if "wait" in kw: wait = kw.wait
        else: wait = None
        self.pushing = True
        while self._status:
            args = self._outqueue.get(timeout=wait)
            if not args: continue
            logging.debug("# IRC/push %s" % str(args))
            if not args or not self._status: break
            try: self.out(*args)
            except: error()
        logging.warn("%s/push end" % self.type)

    def say(self, *args, **kwargs):
        """
            say(channel, txt)

            Say text on channel.

            :param str channel: IRC channel
            :param str txt: text to output to channel

            >>> bot = IRC()
            >>> bot.say("#dunkbots", "yo!") 

        """
        logging.info("# IRC/say %s" % get_location())
        self._outqueue.put(args)
 
    def out(self, *args, **kwargs):
        """
            out(channel, txt)

            Output text on channel, multiple lines if needed. Sleep 3 seconds between sending a string to the server.

            :param str channel: IRC channel
            :param str txt: text to output to channel

            >>> bot = IRC()
            >>> bot.out("#dunkbots", "yo!")

        """
        channel, txt, *rest = args
        txt_list = split_txt(txt)
        for txt in txt_list:
            for t in txt.split("\n"):
                tt = t.strip()
                if not tt: continue
                self.privmsg(channel, tt)
                time.sleep(3.0)

    ## BIND, CONNECT AND LOGON

    def bind(self):
        """ 
            bind(txt)

            Bind the bot to the appropriate socket

            >>> bot = IRC()
            >>> bot.bind()

        """
        server = self.server
        try: self._oldsock.bind((server, 0))
        except socket.error:
            if not server:
                try: socket.inet_pton(socket.AF_INET6, self.server)
                except socket.error: pass
                else: server = self.server
            if not server:
                try: socket.inet_pton(socket.AF_INET, self.server)
                except socket.error: pass
                else: server = self.server
            if not server:
                ips = []
                try:
                    for item in socket.getaddrinfo(self.server, None):
                        if item[0] in [socket.AF_INET, socket.AF_INET6] and item[1] == socket.SOCK_STREAM:
                            ip = item[4][0]
                            if ip not in ips: ips.append(ip)
                except socket.error: pass
                else: server = random.choice(ips)
        return server

    def logon(self, *args, **kwargs):
        """ 
            logon(txt)

            Bind the bot to the appropriate socket

            :param str kwargs["password"]: optional password to connect with
            :param str kwargs["nick"]: nick name of the bot
            :param str kwargs["username"]: username of the bot
            :param str kwargs["server"]: domain name of the server
            :param str kwargs["realname"]: realname  of the bot

            >>> bot = IRC()
            >>> bot.push()

        """
        kw = Object(**kwargs)
        if "password" in kw: self.raw("PASS %s" % kw.password)
        if "nick" in kw: self.nick = kw.nick
        if "username" in kw: self.username = kw.username
        if "server" in kw: self.server = kw.server
        if "realname" in kw: self.realname = kw.realname
        self.raw("NICK %s" % self.nick or "meds")
        self.raw("USER %s localhost %s :%s" % (self.username or "meds", self.server or "localhost", self.realname or "meds"))
        self.ready()

    def connecting(self, *args, **kwargs):
        """
            connecting()

            Try to connect to server until the bot is connected. Sleep 10 seconds between attempts. Logon to server after connect has succeded.

            >>> bot = IRC()
            >>> bot.connecting()

        """
        self._status = "connect"
        res = None
        while not self._error:
            try: self.connect() ; break
            except NotConnected: break
            except: error() ; time.sleep(10)
        if self._error:  return
        time.sleep(5)
        self.logon()
        return res

    ## QUIT

    def exit(self):
        """
            exit()

            Disconnect from the IRC server.

            >>> bot = IRC()
            >>> bot.exit()

        """
        self.quit()

    def close(self):
        """
            close()

            Close the connection to the IRC server.

            :raises meds.errors.NotConnected: Exception raised is bot is not connected

            >>> bot = IRC()
            >>> bot.close()

        """
        if 'ssl' in self and self['ssl']: self.oldsock.shutdown(1) ; self.oldsock.close()
        else: self._sock.shutdown(1) ; self._sock.close()
        self.fsock.close()

    ## IRC COMMANDS

    def donick(self, *args, **kwargs):
        """
            donick(nick)

            Set nick of the bot.

            :param str nick: IRC nick name

            >>> bot = IRC()
            >>> bot.donick("dunker")

        """
        self.raw('NICK %s\n' % args[0][:16])

    def join(self, channel, password=""):
        """
            join(channel, password="")

            Join an IRC channel.

            :param str channel: IRC channel
            :param str password: optional password to use

            >>> bot = IRC()
            >>> bot.join("#dunkbots")

        """
        if password: self.raw('JOIN %s %s' % (channel, password))
        else: self.raw('JOIN %s' % channel)

    def part(self, channel):
        """
            part(channel)

            Part an IRC channel.

            :param str channel: IRC channel

            >>> bot = IRC()
            >>> bot.part("#dunbkots")

        """
        self.raw('PART %s' % channel)
        if channel in self.channels:
            self.channels.remove(channel)
            self.save()

    def who(self, channel):
        """
            who(nick)

            Require information of users on a channel

            :param str channel: channel name

            >>> bot = IRC()
            >>> bot.who("#dunkbots")

        """
        self.raw('WHO %s' % channel)

    def names(self, channel):
        """
            names(channel))

            Show information of users on  a channel

            :param str channel: IRC channel 

            >>> bot = IRC()
            >>> bot.names("#dunkbots")

        """
        self.raw('NAMES %s' % channel)

    def whois(self, nick):
        """
            whois(nick)

            Query a user

            :param str channel: Nick name

            >>> bot = IRC()
            >>> bot.whois("dunker")

        """
        self.raw('WHOIS %s' % nick)

    def privmsg(self, channel, txt):
        """
            privmsg(channel, password="")

            Use PRIVMSG to send text to the server.

            :param str channel: IRC channel
            :param str txt: text

            >>> bot = IRC()
            >>> bot.privmsg("#dunkbots", "yo!")

        """
        self.raw('PRIVMSG %s :%s' % (channel, txt))

    def voice(self, channel, nick):
        """
            voice(channel, nick)

            Join an IRC channel.

            :param str channel: IRC channel
            :param str nick: Nick name

            >>> bot = IRC()
            >>> bot.voice("#dunkbots", "botfather")

        """
        self.raw('MODE %s +v %s' % (channel, nick))

    def doop(self, channel, nick):
        """
            doop(channel, nick)

            Give operator status.

            :param str channel: IRC channel to join
            :param str nick: Nick name

            >>> bot = IRC()
            >>> bot.doop("#dunkbots", "botfather")

        """
        self.raw('MODE %s +o %s' % (channel, nick))

    def delop(self, channel, nick):
        """
            delop(channel, nick)

            Remove operator status.

            :param str channel: IRC channel to join
            :param str nick: Nick name

            >>> bot = IRC()
            >>> bot.delop("#dunkbots", "botfather")

        """
        self.raw('MODE %s -o %s' % (channel, nick))

    def quit(self, reason='https://pikacode.com/bthate/meds2'):
        """
            quit(reason="")

            Quit IRC server

            :param str reason: optional reason to quit

            >>> bot = IRC()
            >>> bot.quit("ltrs")

        """
        self.raw('QUIT :%s' % reason)

    def notice(self, channel, txt):
        """
            notice(channel, txt)

            Send NOTICE to a user.

            :param str channel: IRC channel to join
            :param str txt: text to send

            >>> bot = IRC()
            >>> bot.notice("#dunkbots", "yo!")

        """
        self.raw('NOTICE %s :%s' % (channel, txt))

    def ctcp(self, nick, txt):
        """
            ctcp(nick, txt)

            Send a CTCP to a user.

            :param str nick: Nick name
            :param str txt: text to send

            >>> bot = IRC()
            >>> bot.ctcp("botfather", "version")

        """
        self.raw("PRIVMSG %s :\001%s\001" % (nick, txt))

    def ctcpreply(self, channel, txt):
        """
            ctcpreply(channel, txt)

            Give a reply to a CTCP request given to the bot.

            :param str channel: IRC channel
            :param str txt: text to send

            >>> bot = IRC()
            >>> bot.ctcpreply("#dunkbots", "yo!")

        """
        self.raw("NOTICE %s :\001%s\001" % (channel, txt))

    def action(self, channel, txt):
        """
            action(channel, txt)

            Send a ACTION to a channel.

            :param str channel: IRC channel
            :param str txt: text to send

            >>> bot = IRC()
            >>> bot.action("#dunkbots", "yo!") # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
             ...
            meds.error.NotConnected: localhost

        """
        self.raw("PRIVMSG %s :\001ACTION %s\001" % (channel, txt))

    def getchannelmode(self, channel):
        """
            getchannelmode(channel)

            Query channel modes.

            :param str channel: IRC channel

            >>> bot = IRC()
            >>> bot.getchannelmode("#dunkbots")

        """
        self.raw('MODE %s' % channel)

    def settopic(self, channel, txt):
        """
            settopic(channel, txt)

            Set the topic of a channel.

            :param str channel: IRC channel
            :param str txt: Topic text 

            >>> bot = IRC()
            >>> bot.join("#dunkbots")

        """
        self.raw('TOPIC %s :%s' % (channel, txt))

    def ping(self, *args, **kwargs):
        """
            ping(txt="")

            Give a PING reply.

            :param str txt: optional PING response.

            >>> bot = IRC()
            >>> bot.ping("yo!")

        """
        self.raw('PING :%s' % args[0])

    def pong(self, *args, **kwargs):
        """
            pong(txt="")

            Give a PONG reply.

            :param str txt: optional reply text

            >>> bot = IRC()
            >>> bot.pong("yo!")

        """
        self.raw('PONG :%s' % args[0])

def init(*args, **kwargs):
    """
        init(*args, **kwargs)

        meds.bots.irc initialisation code

    """
    ncfg = main.last("cfg", "irc")
    if not ncfg: ncfg = cfg.irc ; ncfg.save()
    bot = IRC(**ncfg)
    fleet.append(bot)
    main.put(bot.start)
    main.put(bot.push)

def stop(*args, **kwargs):
    """
        stop(*args, **kwargs)

        meds.bots.irc shutdown code

    """
    for bot in fleet:
       if bot.type == "irc": bot.exit()

## POSTFIXCOUNT - how many arguments used in a Anderen message

pfc = {}
pfc['NICK'] = 0
pfc['QUIT'] = 0
pfc['SQUIT'] = 1
pfc['JOIN'] = 0
pfc['PART'] = 1
pfc['TOPIC'] = 1
pfc['KICK'] = 2
pfc['PRIVMSG'] = 1
pfc['NOTICE'] = 1
pfc['SQUERY'] = 1
pfc['PING'] = 0
pfc['ERROR'] = 0
pfc['AWAY'] = 0
pfc['WALLOPS'] = 0
pfc['INVITE'] = 1
pfc['001'] = 1
pfc['002'] = 1
pfc['003'] = 1
pfc['004'] = 4
pfc['005'] = 15
pfc['302'] = 1
pfc['303'] = 1
pfc['301'] = 2
pfc['305'] = 1
pfc['306'] = 1
pfc['311'] = 5
pfc['312'] = 3
pfc['313'] = 2
pfc['317'] = 3
pfc['318'] = 2
pfc['319'] = 2
pfc['314'] = 5
pfc['369'] = 2
pfc['322'] = 3
pfc['323'] = 1
pfc['325'] = 3
pfc['324'] = 4
pfc['331'] = 2
pfc['332'] = 2
pfc['341'] = 3
pfc['342'] = 2
pfc['346'] = 3
pfc['347'] = 2
pfc['348'] = 3
pfc['349'] = 2
pfc['351'] = 3
pfc['352'] = 7
pfc['315'] = 2
pfc['353'] = 3
pfc['366'] = 2
pfc['364'] = 3
pfc['365'] = 2
pfc['367'] = 2
pfc['368'] = 2
pfc['371'] = 1
pfc['374'] = 1
pfc['375'] = 1
pfc['372'] = 1
pfc['376'] = 1
pfc['378'] = 2
pfc['381'] = 1
pfc['382'] = 2
pfc['383'] = 5
pfc['391'] = 2
pfc['392'] = 1
pfc['393'] = 1
pfc['394'] = 1
pfc['395'] = 1
pfc['262'] = 3
pfc['242'] = 1
pfc['235'] = 3
pfc['250'] = 1
pfc['251'] = 1
pfc['252'] = 2
pfc['253'] = 2
pfc['254'] = 2
pfc['255'] = 1
pfc['256'] = 2
pfc['257'] = 1
pfc['258'] = 1
pfc['259'] = 1
pfc['263'] = 2
pfc['265'] = 1
pfc['266'] = 1
pfc['401'] = 2
pfc['402'] = 2
pfc['403'] = 2
pfc['404'] = 2
pfc['405'] = 2
pfc['406'] = 2
pfc['407'] = 2
pfc['408'] = 2
pfc['409'] = 1
pfc['411'] = 1
pfc['412'] = 1
pfc['413'] = 2
pfc['414'] = 2
pfc['415'] = 2
pfc['421'] = 2
pfc['422'] = 1
pfc['423'] = 2
pfc['424'] = 1
pfc['431'] = 1
pfc['432'] = 2
pfc['433'] = 2
pfc['436'] = 2
pfc['437'] = 2
pfc['441'] = 3
pfc['442'] = 2
pfc['443'] = 3
pfc['444'] = 2
pfc['445'] = 1
pfc['446'] = 1
pfc['451'] = 1
pfc['461'] = 2
pfc['462'] = 1
pfc['463'] = 1
pfc['464'] = 1
pfc['465'] = 1
pfc['467'] = 2
pfc['471'] = 2
pfc['472'] = 2
pfc['473'] = 2
pfc['474'] = 2
pfc['475'] = 2
pfc['476'] = 2
pfc['477'] = 2
pfc['478'] = 3
pfc['481'] = 1
pfc['482'] = 2
pfc['483'] = 1
pfc['484'] = 1
pfc['485'] = 1
pfc['491'] = 1
pfc['501'] = 1
pfc['502'] = 1
pfc['700'] = 2
