# Object.py
#
#

""" Core Object. A dict with dotted named access. """

__copyright__ = "Copyright 2015, B.H.J Thate"

## IMPORTS

from meds.utils.file import list_files
from meds.utils.parse import txt_parse
from meds.utils.time import fn_time, rtime
from meds.utils.name import typed, name
from meds.utils.lock import locked
from meds.errors import NoJSON

import threading
import hashlib
import logging
import errno
import fcntl
import types
import json
import time
import os


## OBJECT

class Object(dict):

    """ Object on which the rest of the program is based. """

    def __getattribute__(self, name):
        if name not in self: return super().__getattribute__(name)
        return self[str(name)]

    def __getitem__(self, name):
        try: return super().__getitem__(int(name))
        except: return super().__getitem__(name)
         
    def __getattr__(self, name):
        if name == "url": return self.urled()
        if name == "_ready":  self._ready = threading.Event()
        if name == "skip": return ["parsed", ] 
        if name == "cc": return "!"
        if name == "type": return typed(self)
        if name not in self: raise AttributeError(name)
        return self[name]

    def __setattr__(self, name, value):
        if name in self and type(self[name]) in [types.FunctionType, types.MethodType]: raise WrongType(name)
        return dict.__setitem__(self, name, value)

    def __contains__(self, name):    
        try: self[name] ; return True
        except KeyError: return False

    ## OUTPUT

    def reply(self, *args, **kwargs):
        """
            reply(txt)

            Send a reply to orginiating channel.

            >>> o = Object()
            >>> o.reply("hello")
            hello
                   
        """
        if "outer" in self: self.outer.write(str(args[0]) + "\n") ; self.outer.flush() ; return
        if "channel" in self: self.say(self.channel, args[0])
        elif "origin" in self: self.say(self.origin, args[0])
        else: self.say("", args[0])

    def say(self, *args, **kwargs):
        """
            say(channel, txt)

            Say text on a channel. 

            >>> o = Object()
            >>> o.say("#testchannel", "hello")
            hello

        """
        if "_target" in self: self._target.say(*args, **kwargs)
        else: print(args[1])

    def ok(self, *args, **kwargs):
        """
            ok(txt="")

            Reply with 'ok' and some descriptive text if provided. 

            >>> o = Object()
            >>> o.ok("done")
            ok done
                    
        """
        txt = " ".join([str(a) for a in args])
        self.reply("ok %s" % txt)

    def show(self, *args, **kwargs):
        """
            show()

            Return the data of the Object in key=value pairs

            >>> o = Object()
            >>> o.bla = "mekker"
            >>> o.show()
            ['bla=mekker']
                   
        """
        return ["%s=%s" % (a, self[a]) for a in sorted(self.names())]

    def display(self, *args, **kwargs):
        """
            display(args)

            Return a string constructed based on a list of keys. Include days ago as well.

            >>> o = Object()
            >>> o.txt = "test"
            >>> o.date = time.ctime(time.time())
            >>> o.display(["txt", ])
            'test - '
                   
        """
        keys = args[0]
        txt = " ".join([str(getattr(self, str(key), "")) for key in keys])
        txt = txt.rstrip() 
        txt += " - %s" % self.days()
        return txt 

    def announce(self, *args, **kwargs):
        """
            announce(txt)

            Announce a text on all channels of all bots.

            >>> from meds.bots import TestBot
            >>> b = TestBot()
            >>> b.announce("hello")
                   
        """
        from meds.main import fleet
        thrs = []
        for bot in fleet: thrs.append(self.put(bot.announce, *args, **kwargs))
        if thrs: self.collect(thrs)
        return thrs

    ## SELECTOR

    def all(self, *args, **kwargs):
        """
            all(*args)

            Select all records based on a optional prefix.

            >>> o = Object()
            >>> res = o.all("cfg", "irc")
                   
        """
        a = sorted(list_files(self.path(), *args, **kwargs), key=lambda x: fn_time(x))
        return a

    def selected(self, *args, **kwargs):
        """

            selected(parsed)

            Select all records based on a parsed object.

            >>> o = Object()
            >>> p = txt_parse(o, "find From=.nl time=2012-09 Subject")
            >>> res = o.selected(p)

        """
        parsed = args[0]
        objs = []
        uniqlist = []
        for fn in self.all(parsed.args[0], **parsed.want):
            try: obj = Object().load(fn)
            except: logging.warn("fail %s" % fn) ; continue
            if "deleted" in obj and obj.deleted: continue
            if not obj.selector(parsed.args): continue
            if not obj.wanted(parsed.want): continue
            if obj.notwanted(parsed.notwant): continue
            uniq = parsed.switch.get("uniq", "")
            if uniq:
                val = obj.get(uniq, "")
                if val in uniqlist: continue
                else: uniqlist.append(val)
            objs.append(obj)
        return objs
 
    def selector(self, *args, **kwargs):
        """ 
            selector(args)

            See if this objects has the desired attributes. 

            >>> o = Object()
            >>> p = txt_parse(o, "find From=.nl time=2012-09 Subject")
            >>> o.selector(p.args)
            False
                   
        """
        keys = args[0]
        if not keys: return False
        for key in keys:
            if key == "time": continue
            if key in self: return True
            if not getattr(self, key, None): return False
        return True

    def wanted(self, *args, **kwargs):
        """ 
            wanted(parsed)

            See if this objects has the desired attributes with their desired values.

            >>> from meds.bots import Bot
            >>> bot = Bot()
            >>> o = txt_parse(bot, "find From=.nl time=2012-09 Subject")
            >>> o.wanted(o.want)
            True
                   
        """
        want = args[0]
        go = True
        for key, value in want.items():
           if key == "format": continue
           if key not in self: continue
           if value.startswith("-"): continue
           if value not in str(self[key]): go = False ; break 
        return go

    def notwanted(self, *args, **kwargs):
        """ 
            notwanted(parsed)

            See if this object doesn't have not desired key/values.

            >>> from meds.bots import Bot
            >>> bot = Bot()
            >>> p = txt_parse(bot, "find From=.nl time=2012-09 Subject")
            >>> bot.notwanted(p.want)
            False
                   
        """
        notwant = args[0]
        for key, value in notwant.items():
           if key == "format": continue
           if key not in self: continue
           if value in self[key]: return True
        return False

    def allowed(self, *args, **kwargs):
        """ 

            allowed(userhost)

            Method to see if a use to allowed based of origin of that user.

            >>> o = Object()
            >>> o.allowed("root@cli")
            False
                   
        """
        origin = args[0]
        for fn in self.all("user"):
            obj = Object().load(fn)
            if "user" in obj and origin == obj.user: return True
        return False

    ## FLEET

    def search(self, *args, **kwargs):
        """ 
            search(txt)

            Search all attribute names containing substring txt. 

            >>> object = Object()
            >>> object.data = "test"
            >>> object.search("da")
            ['data']

        """
        l = []
        for key in self.names():
            if args[0] in key: l.append(key)
        return l

    ## STATE

    def register(self, *args, **kwargs):
        """ 
            register(cmd, func)

            Register a command, function combination

            >>> object = Object()
            >>> def test(event): event.reply("yooo")
            >>> object.register("test", test)
            >>> object["test"][0].cmnd
            'test'

        """
        from meds.utils.trace import get_plugname
        o = Object()
        o.cmnd = args[0]
        o.func = args[1]
        o.plugname = get_plugname()
        if o.cmnd not in self: self[o.cmnd] = []
        todo = []
        for obj in self[o.cmnd]:
            if obj.cmnd == o.cmnd and obj.plugname == o.plugname: todo.append(obj)
        for obj in todo: self[o.cmnd].remove(obj) ; logging.info("! remove %s" % name(obj))
        self[o.cmnd].append(o)
        logging.info("! %s/register %s" % (self.type, o.cmnd))

    ## GETTERS

    def match(self, *args, **kwargs):
        """
            match(cmnd)

            Match all object based on the last value of their dotted keynames.

            >>> object = Object()
            >>> object.match("ls") # doctest: +ELLIPSIS
            [...]

        """
        rest = args[0]
        result = []
        for x in set(self.search(rest)):
            if x.split(".")[-1] != rest: continue
            obj = self.get(x)
            if type(obj) in [list, tuple]: result.extend(obj) ; continue
            result.append(obj)
        return result

    ## TIME RELATED

    def days(self, *args, **kwargs):
        """ 
            days()

            Get the number of days relative to the ding's creation time.

            >>> object = Object()
            >>> object.Date = time.ctime(time.time())
            >>> object.days()
            ''

        """
        from meds.utils.time import elapsed
        t1 = time.time()
        t2 = self.timed()
        if t2:
            time_diff = float(t1 - t2)
            return elapsed(time_diff)

    def dated(self, *args, **kwargs):
        """ 
            dated()

            Retrieve the creation time of this object.

            >>> object = Object()
            >>> object.added = 'Sun Aug 23 16:07:16 2015'
            >>> object.dated()
            'Sun Aug 23 16:07:16 2015'

        """
        val = ""
        if "Date" in self: val = self.Date
        elif "date" in self: val = self.date
        elif "published" in self: val = self.published
        elif "added" in self: val = self.added
        elif "saved" in self: val = self.saved
        elif "timed" in self: val = self.timed
        return val

    def timed(self, *args, **kwargs):
        """ 
            timed()

            Retrieve the time stored in the path name of this object.

            >>> object = Object()
            >>> object.added = 'Sun Aug 23 16:07:16 2015'
            >>> object.timed()
            1440338836.0

        """
        from meds.utils.time import short_date, to_time
        t = short_date(self.dated())
        if t: t = to_time(t)
        if not t and "_path" in self: t = fn_time(self._path)
        if not t: t = fn_time(rtime())
        return t

    def names(self, *args, **kwargs):
        """ 
            names()

            Provide a list of keys suitable for iteration, skipping keys starting with a underscore.

            >>> object = Object()
            >>> object.bla = "mekker"
            >>> list(object.names())
            ['bla']

        """
        for key in self.keys():
            k = str(key)
            if k.startswith("_"): continue
            yield k

    def slice(self, *args, **kwargs):
        """ 
            slice(args)

            Provide a sub directory based of a optional provided key list.

            >>> object = Object()
            >>> object["bla"] = 1
            >>> object["bla2"] = 2
            >>> object.slice(["bla"])
            {'bla': 1}

        """
        o = Object()
        try: arguments = args[0]
        except IndexError: arguments = self.names()
        for arg in arguments:
            try: o[arg] = self[arg]
            except KeyError: continue
        return o

    def urled(self, *args, **kwargs):
        """ 
            urled()

            Url of the Object's file so that it can be retrieved when API server is running.

            >>> object = Object()
            >>> object._path = "bla"
            >>> object.urled()
            'http://wakker:10102/bla'

        """
        from meds.cfg import cfg
        if "_path" in self: fn = os.sep.join(self._path.split(os.sep)[-2:])
        else: fn = ""
        return "http://%s:%s/%s" % (cfg.rest.hostname, cfg.main.port, fn)

    def root(self, *args, **kwargs):
        """ 
            root()

            Return the working directory.

            >>> object = Object()
            >>> object.root()
            '/home/bart/.meds'

        """
        from meds.utils.file import check_permissions
        from meds.cfg import cfg
        if "workdir" in self: root = self.workdir
        else: root = cfg.main.workdir
        path = os.path.abspath(root)
        check_permissions(path)
        return path

    def path(self, *args, **kwargs):
        """ 
            path()

            Return the path to the object.

            >>> object = Object()
            >>> object.prefix = "cfg"
            >>> object.path()
            '/home/bart/.meds/cfg'

        """
        from meds.utils.join import j
        root = self.root()
        if "prefix" in self: root = j(root, self.prefix)
        return os.path.abspath(root) 

    ## INPUT

    def load(self, *args, **kwargs):
        """ 
            load(path="")

            Load a JSON file..

            >>> object = Object()
            >>> object.load("bla")
            {'_path': 'bla'}

        """
        if args: path = args[0]
        else: path = self._path
        ondisk = self.read(path)
        fromdisk = json.loads(ondisk)
        if "data" in fromdisk: self.update(fromdisk["data"])
        else: self.update(fromdisk)
        if "saved" in fromdisk: self.saved = fromdisk["saved"]
        self._path = path
        logging.debug("load %s" % path)
        return self

    def read(self, *args, **kwargs):
        """ 
            read(path)

            Read a JSON file.

            >>> object = Object()
            >>> object.read("bla")
            '{}'

        """
        path = args[0]
        try: f = open(path, "r", encoding="utf-8")
        except IOError as ex:
            if ex.errno == errno.ENOENT: return "{}"
            raise
        res = ""
        for line in f:
            if not line.strip().startswith("#"): res += line
        if not res.strip(): return "{}"
        f.close()
        return res

    ## PERSISTENCE

    def prepare(self, *args, **kwargs):
        """ 
            prepare()

            Create JSON string.

            >>> object = Object()
            >>> s = object.prepare()

        """
        from meds.utils.trace import get_plugname
        todisk = Object()
        todisk.data = self.slice()
        todisk.saved_from = get_plugname(2)
        todisk.type = self.type
        todisk.saved = self.saved = time.ctime(time.time())
        todisk.signature = todisk.data.signature()
        try: result = todisk.json(indent=4, ensure_ascii=False, sort_keys=True)
        except TypeError: raise NoJSON()
        return result

    def save(self, *args, **kwargs):
        """ 
            save(fname="")

            Save as JSON to disk. Use a timestamp if no argument provided.

            >>> object = Object()
            >>> object.save("bla")
            'bla'

        """
        from meds.utils.time import rtime
        from meds.utils.join import j
        if not args: t = rtime()
        else: t = args[0]
        self.sync(j(self.path(), t))
        return t

    @locked
    def sync(self, *args, **kwargs):
        """ 
            sync(path="")

            Sync to disk using the current path or use a timestamp.

            >>> object = Object()
            >>> object.prefix = "test"
            >>> path = object.sync()
            >>> '/home/bart/.meds/test/' in path
            True

        """
        from meds.utils.other import headertxt
        from meds.utils.time import rtime
        from meds.utils.file import cdir
        from meds.utils.join import j
        try: path = args[0]
        except IndexError:
            try: path = self._path
            except AttributeError: path = self._path = j(self.path(), rtime())
        logging.info("! sync %s" % path)
        d, fn = os.path.split(path)
        cdir(d)
        todisk = self.prepare(**kwargs)
        datafile = open(os.path.abspath(path) + ".tmp", 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % "%s characters" % len(todisk))
        datafile.write(todisk)
        datafile.write("\n")
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        os.rename(path + ".tmp", path)
        return path


    ## ITERATORS

    def ls(self, *args, **kwargs):
        """
            ls(*args)

            Show listing of work directory.

            >>> o = Object()
            >>> o.ls("cfg") # doctest: +ELLIPSIS
            [...]

        """
        from meds.cfg import cfg
        fns = self.all(*args)
        prefixes = []
        for fn in fns:
            workdir, path = fn.split(cfg.main.workdir) 
            boh, prefix, *rest = path.split(os.sep)
            if path and not path in prefixes: prefixes.append(path)
        return prefixes

    def prefixes(self, *args, **kwargs):
        """
            prefixes()

            Show prefixes of work directory.

            >>> o = Object()
            >>> o.prefixes() # doctest: +ELLIPSIS
            [...]

        """
        from meds.cfg import cfg
        fns = self.all(*args)
        prefixes = []
        for fn in fns:
            workdir, path = fn.split(cfg.main.workdir) 
            boh, prefix, *rest = path.split(os.sep)
            if prefix not in prefixes: prefixes.append(prefix)
        return prefixes

    def first(self, *args, **kwargs):
        """ 
            first(prefix, match="")

            Find first object that has prefix. String to match with value is optional.

            >>> object = Object()
            >>> o = object.first("cfg")
            >>> type(o)
            <class 'meds.object.Object'>

        """

        fns = self.all(args[0])
        for fn in fns:
            obj = Object().load(fn)
            if "deleted" in obj and obj.deleted: continue
            if args and len(args) > 1 and args[1] != obj.get(args[0], ""): continue
            return obj

    def last(self, *args, **kwargs):
        """ 
            last(prefix, match="")

            Find last object that has prefix. String is match with value is optional.

            >>> object = Object()
            >>> o = object.last("cfg")
            >>> type(o)
            <class 'meds.object.Object'>

        """
        fns = self.all(args[0])[::-1]
        res = None
        for fn in fns[::-1]:
            obj = Object().load(fn)
            if "deleted" in obj and obj.deleted: continue
            if args and len(args) > 1 and args[1] != obj.get(args[0], ""): continue
            res = obj
        return res

    ## WAITERS

    def ready(self):
        """ 
            ready()

            Signal to ready state. 

            >>> o = Object()
            >>> o.ready()
  
        """
        logging.debug("! %s/ready" % self.type)
        self._ready.set()

    def clear(self):
        """
            clear()

            Clear the ready state. 

            >>> o = Object()
            >>> o.clear()

        """
        logging.debug("! %s/clean" % self.type)
        self._ready.clear()

    def wait(self, sec=180.0):
        """
            wait(sec=180.0)

            Wait for ready state. 

            >>> o = Object()
            >>> o.wait(1.0)

        """
        logging.debug("! %s/wait" % self.type)
        self._ready.wait(sec)

    ## JSON

    def full(self, *args, **kwargs):
        """ 
            full()

            Create full JSON from this Object

            >>> object = Object()
            >>> object.bla = "mekker"
            >>> object.full()
            '{"bla": "mekker"}'

        """
        from meds.utils.serialize import smooth
        return json.dumps(self, default=smooth, *args, **kwargs)

    def json(self, *args, **kwargs):
        """ 
            json()

            Create JSON string.

            >>> object = Object()
            >>> object.bla = "mekker"
            >>> object.json()
            '{"bla": "mekker"}'

        """
        from meds.utils.serialize import smooth
        return json.dumps(self.slice(*args, **kwargs), default=smooth, *args, **kwargs)

    def pretty(self, *args, **kwargs):
        """ 
            pretty()

            Make a pretty JSON string.

            >>> object = Object()
            >>> object.bla = "mekker"
            >>> s = object.pretty()
            >>> type(s)
            <class 'str'>

        """
        from meds.utils.serialize import pretty
        return json.dumps(self.slice(), indent=4, default=pretty, sort_keys=True, ensure_ascii=True)

    ## HELPERS

    def signature(self):
        """ 
            signature()

            Create a signature of this Object.

            >>> object = Object()
            >>> object.signature()
            'bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f'

        """
        return str(hashlib.sha1(bytes(str(self), "utf-8")).hexdigest())


## OBJECT OF OBJECTS

class OO(list): pass

